package redis

import (
	"encoding/json"
	"fmt"
	"github.com/go-redis/redis"
	"github.com/open-io/oio-exporter/common"
	"log"
	"regexp"
	"strconv"
	"strings"
	"time"
)

var scriptGetAccounts = redis.NewScript(`
    return redis.call("hgetall", "accounts:")
`)

var scriptGetContCount = redis.NewScript(`
    local key = "containers:" .. KEYS[1]
    return redis.call("ZCOUNT", key, 0, 0)
`)

var scriptAcctInfo = redis.NewScript(`
	local accts = redis.call("hgetall", "accounts:")
	local res = {}
	local index = 0
	for i, acc in ipairs(accts) do
        if i % 2 == 1 then
                local acct_key = "account:" .. acc;
                res[index] = {acc, redis.call('HGET', acct_key, 'bytes'), redis.call("HGET", acct_key, 'objects')}
                index = index + 1
        end;
	end;
	return cjson.encode(res);
`)

type client struct {
	Addr       string
	Client     *redis.Client
	IsAccount  bool
	IsCH       bool
	IsBucketDB bool
}

type daemon struct {
	clients    []*client
	interval   time.Duration
	lastUpdate time.Time
	services   **map[string][]*common.Service
}

var whitelist = map[string]bool{
	"used_memory":                 true,
	"used_memory_rss":             true,
	"used_memory_lua":             true,
	"mem_fragmentation_ratio":     true,
	"rdb_changes_since_last_save": true,
	"total_connections_received":  true,
	"total_commands_processed":    true,
	"instantaneous_ops_per_sec":   true,
	"total_net_input_bytes":       true,
	"total_net_output_bytes":      true,
	"keyspace_hits":               true,
	"keyspace_misses":             true,
	"role":                        true,
	"connected_slaves":            true,
	"repl_backlog_size":           true,
	"db0":                         true,
}

var keysRegexp = regexp.MustCompile(`keys=(\d+)`)

func MakeDaemon(services **map[string][]*common.Service, interval time.Duration) *daemon {
	d := daemon{
		services:   services,
		clients:    []*client{},
		interval:   interval,
		lastUpdate: time.Unix(0, 0),
	}
	return &d
}

func (d *daemon) reload(lastUpdate time.Time) {
	log.Println("INFO", "redis module", "reloading config")
	clients := []*client{}
	if redisInstances, ok := (**d.services)["redis"]; ok && len(redisInstances) > 0 {
		for _, svc := range redisInstances {
			target := fmt.Sprintf("%s:%d", svc.IP, svc.Port)
			newClient := client{
				Addr: target,
				Client: redis.NewClient(&redis.Options{
					Addr: target,
				}),
			}
			val, err := newClient.Client.HLen("accounts:").Result()
			if err == nil && val > 0 {
				newClient.IsAccount = true
			}
			clients = append(clients, &newClient)
		}
	}
	d.clients = clients
	d.lastUpdate = lastUpdate
}

func (d *daemon) Run(m *common.MetricData, lastUpdate **time.Time) {
	for {
		if (*lastUpdate).Sub(d.lastUpdate) > 0 {
			d.reload(**lastUpdate)
		}
		for _, c := range d.clients {
			if c.IsAccount {
				d.collectAccount(c, m)
			}
			d.collectStats(c, m)
		}
		time.Sleep(d.interval)
	}
}

func (d *daemon) collectStats(c *client, m *common.MetricData) {
	data, err := c.Client.Info().Result()
	if err != nil {
		m.IncrError("redis", "redis")
		log.Println("WARN", "when collecting redis stats", c, err)
		return
	}
	for _, line := range strings.Split(data, "\r\n") {
		kv := strings.Split(line, ":")
		if len(kv) != 2 {
			continue
		}
		if _, ok := whitelist[kv[0]]; ok {
			// Match keys in db entry
			if strings.HasPrefix(kv[0], "db") {
				keys := keysRegexp.FindStringSubmatch(kv[1])
				if len(keys) > 1 {
					valKeys, err := strconv.ParseFloat(keys[1], 64)
					if err != nil {
						m.IncrError("redis", "redis")
						log.Println("WARN", "when collecting redis stats", c, err)
						continue
					}
					m.Set(&common.Metric{
						Name: "redis_keys",
						Labels: [][]string{
							[]string{"addr", c.Addr},
						},
						Value: valKeys,
					})
				}
				// Format role:master or role:slave
			} else if kv[0] == "role" {
				if kv[1] == "master" {
					m.Set(&common.Metric{
						Name: "redis_is_master",
						Labels: [][]string{
							[]string{"addr", c.Addr},
						},
						Value: float64(1),
					})
				} else {
					m.Set(&common.Metric{
						Name: "redis_is_master",
						Labels: [][]string{
							[]string{"addr", c.Addr},
						},
						Value: float64(1),
					})
					m.Set(&common.Metric{
						Name: "connected_slaves",
						Labels: [][]string{
							[]string{"addr", c.Addr},
						},
						Value: float64(0),
					})
				}
			} else {
				val, err := strconv.ParseFloat(kv[1], 64)
				if err != nil {
					m.IncrError("redis", "redis")
					log.Println("WARN", "when collecting redis stats", c, err)
					continue
				}
				m.Set(&common.Metric{
					Name: "redis_" + kv[0],
					Labels: [][]string{
						[]string{"addr", c.Addr},
					},
					Value: val,
				})
			}
		}
	}
}

func (d *daemon) collectAccount(c *client, m *common.MetricData) {
	acctInfo, err := scriptAcctInfo.Run(c.Client, []string{}, 0).Result()
	if err != nil {
		log.Println("WARN", "when collecting account info", c, err)
		return
	}
	acctObj := map[string][]string{}
	err = json.Unmarshal([]byte(acctInfo.(string)), &acctObj)
	if err != nil {
		m.IncrError("redis", "account")
		log.Println("WARN", "when collecting account info", c, err)
		return
	}
	for _, data := range acctObj {
		valKiB, err := strconv.ParseFloat(data[1], 64)
		if err != nil {
			m.IncrError("redis", "account")
			log.Println("WARN", "when collecting account objects", c, err)
			continue
		}
		valObj, err := strconv.ParseFloat(data[2], 64)
		if err != nil {
			m.IncrError("redis", "account")
			log.Println("WARN", "when collecting account objects", c, err)
			continue
		}
		m.Set(&common.Metric{
			Name: "account_kiB",
			Labels: [][]string{
				[]string{"account", data[0]},
			},
			Value: valKiB / 1024,
		})
		m.Set(&common.Metric{
			Name: "account_objects",
			Labels: [][]string{
				[]string{"account", data[0]},
			},
			Value: valObj,
		})
	}

	accounts, err := scriptGetAccounts.Run(c.Client, []string{}, 0).Result()
	if err != nil {
		m.IncrError("redis", "account")
		log.Println("WARN", "when listing accounts", c, err)
		return
	}
	for _, acct := range accounts.([]interface{}) {
		if acct == "1" {
			continue
		}
		count, err := scriptGetContCount.Run(c.Client, []string{acct.(string)}, 1).Result()
		if err != nil {
			m.IncrError("redis", "account")
			log.Println("WARN", "when collecting account containers", acct, err)
			continue
		}
		m.Set(&common.Metric{
			Name: "account_containers",
			Labels: [][]string{
				[]string{"addr", c.Addr},
				[]string{"account", acct.(string)},
			},
			Value: float64(count.(int64)),
		})
	}
}
